/************************************************************************
    frap/defrap - fractal image compressor/decompressor
    Copyright (C) 1997-2000 Matthias Ruhl

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

    The author can be contacted at:
      MIT Laboratory for Computer Science
      200 Technology Square
      Cambridge, MA 02139
      USA
    or via e-mail: ruhl@theory.lcs.mit.edu
*************************************************************************/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "fileio.h"

static const int minval = 1<<WBITS;

static FILE *f = NULL;
static int enc = 0;

int packpos;

int unpack(int bits, FILE *inp)
{
  static int ch,ex,ret;

  ex = 1; ret = 0;
  while(bits--)
    {
      if(packpos%8 == 0)
	ch = fgetc(inp);
      if(ch & 1) ret |= ex;
      ex <<= 1; ch >>= 1;
      packpos++;
    }
  return ret;
}

void pack(int val, int bits, FILE *out)
{
  static int buf = 0;

  while(bits--)
    {
      buf |= (val & 1)<<(packpos%8);
      val >>= 1;
      packpos++;
      if(packpos%8 == 0)
	{
	  fputc(buf,out); buf = 0;
	}
    }
}

void arith_init(int ee, FILE *ff)
{
  enc = ee; f = ff;
}

static int mdiv(int a, int b, int c)
{
  return (int) (((double)a*(double)b)/(double)c + 0.5);
}

static void lshift(bigint b)
{
  int i;

  for(i=BIGINTLEN-1;i>=0;i--)
    {
      b[i] <<= 1;
      if(b[i] & (1<<BIGINTBITS))
	{
	  b[i] ^= (1<<BIGINTBITS);
	  b[i+1] |= 1;
	}
    }
}

static void add(bigint b, int a)
{
  int i;

  b[0] += a;
  for(i=0;i<BIGINTLEN && (b[i] & (1<<BIGINTBITS));i++)
    {
      b[i] ^= (1<<BIGINTBITS);
      b[i+1]++;
    }
}

static int sub(bigint a, bigint b)
{
  int i,tmp;

  for(tmp=0,i=BIGINTLEN-1;i>=0;i--)
    {
      if(tmp)
	{
	  tmp = ((1<<BIGINTBITS) + a[i]) - b[i];
	}
      else tmp = a[i] - b[i];
    }
  return tmp;
}

static int and(bigint a, int pos)
{
  return a[pos/BIGINTBITS] & (1<<(pos%BIGINTBITS));
}

static void erase(bigint b, int pos)
{
  b[pos/BIGINTBITS] ^= (1<<(pos%BIGINTBITS));
}

static void readcode(bigint b, int n)
{
  while(n--)
    {
      lshift(b);
      if(unpack(1,f)) b[0] |= 1;
    }
}

void arith_code(char *c, int len, int *freq)
{
  int cfreq[SYMBOLS+1];
  int i,nextc,cnt=0;
  int lft,lo,hi,diff;
  bigint lwr,upr,code;

  if(enc) /* encoding */
    {
      cfreq[0] = 0;
      for(i=0;i<SYMBOLS;i++) cfreq[i+1] = cfreq[i] + max(1,freq[i]);
    }
  else /* decoding */
    {
      cfreq[0] = 0;
      for(i=0;i<SYMBOLS;i++) cfreq[i+1] = cfreq[i] + freq[i];
    }

  for(i=0;i<BIGINTLEN;i++)
    lwr[i] = upr[i] = code[i] = 0;
  upr[0] = 2*minval-1; lft = WBITS;
  if(!enc) readcode(code,WBITS+1);

  for(i=0;i<len;i++)
    {
      /* if decoding: decode next symbol */
      if(!enc)
	{
	  diff = sub(code,lwr);
	  for(nextc=0;nextc<SYMBOLS;nextc++)
	    if(mdiv(sub(upr,lwr),cfreq[nextc],cfreq[SYMBOLS]) <= diff && 
	       mdiv(sub(upr,lwr),cfreq[nextc+1],cfreq[SYMBOLS]) > diff) break;
	  c[i] = nextc;
	}

      /* update interval boudaries */

      lo = mdiv(sub(upr,lwr),cfreq[(int)c[i]],cfreq[SYMBOLS]);
      hi = mdiv(sub(upr,lwr),cfreq[(int)c[i]+1],cfreq[SYMBOLS]);
      memcpy(upr,lwr,sizeof(bigint)); add(upr,hi); add(lwr,lo);

      /* output all leading digits in lwr and upr that are the same */

      while(lft >= 0 && (and(upr,lft) == and(lwr,lft)))
	{
	  cnt++;
	  if(and(lwr,lft))
	    {
	      if(enc) pack(1,1,f);
	      erase(lwr,lft);
	      erase(upr,lft);
	      if(!enc) erase(code,lft);
	    }
	  else
	    if(enc) pack(0,1,f);
	  lft--;
	}

      /* left-shift lwr und upr until the interval between them is large
	 enough */
      
      if(i != len-1)
	for(diff=sub(upr,lwr);diff<minval;diff<<=1)
	  {
	    if(lft >= BIGINTLEN*BIGINTBITS) 
	      { printf("ERROR acode.c: BIGINTLEN too short!\n"); exit(1); }
	    lshift(lwr); lshift(upr); if(!enc) readcode(code,1);
	    lft++;
	  }
    }
  if(enc)
    /* output the trailing digits (approx. WBITS many) */
    while(lft >= 0)
      {
	cnt++;
	if(and(lwr,lft)) pack(1,1,f); else pack(0,1,f); 
	lft--;
      }
}
